#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "types.h"
#include "defaults.h"
#include "warning.h"
#include "mystring.h"
#include "color.h"
#include "filer.h"
#include "map.h"

float *mapa_z = NULL;
kq_submap *prvsubmap = NULL;
kq_submap *bgmap = NULL;
kq_submap *prvhidsm = NULL;
char autoreloadmap = 0;

static kq_color **allocpole (int sirka, int vyska)
{
  return (kq_color **)malloc(sizeof(kq_color **)*sirka*vyska);
}

kq_map nactimapu (char *filename)
{
  kq_filer *soubor;
  int znak, ch, i, j, k, pocetbarev;
  kq_color *pom, *pomo;
  kq_map vysledek;
  char *fullname;

  for(pom = prvbarva; pom != NULL; pom = pom->next) pom->znak = 0; // pronuluji znaky barev

  fullname = joinstrings(DATADIR, MAPDIR, filename, MAPSUFFIX, NULL);

  soubor = mf_otevri(fullname);
  mf_musibyt(soubor, "/* XPM */");
  mf_najdi(soubor, '"');

  vysledek.sirka = mf_cislo(soubor);
  mf_musibyt(soubor, " ");
  vysledek.vyska = mf_cislo(soubor);
  mf_musibyt(soubor, " ");
  pocetbarev = mf_cislo(soubor);
  mf_musibyt(soubor, " 1\"");

  // nacitam znaky barev:

  for(i=0; i<pocetbarev; i++){
    mf_najdi(soubor, '"');
    znak=mf_znak(soubor);
    mf_musibyt(soubor, "	c ");

    ch=1;
    pom = prvbarva;
    for(j=0; ch; j++){
      ch = mf_znak(soubor);
      if(ch == EOF){
	warning ("Konec souboru pri nacitani %d. barvy", j);
	mf_error(soubor);
      }
      if(ch == '"'){
	ch=0;
	soubor->ch = '"';
      }

      k=j;
      for(pomo = pom; pom; pom = pom->next){
	while(k < j && pom->nazev[k] == pomo->nazev[k]) k++;
	if(k==j && pom->nazev[k] == ch) break;
	k=0;
      }
      if(pom == NULL) break;
    }
    if(pom != NULL) pom->znak = znak;
    mf_najdi(soubor, '"');
  }

  vysledek.pole = allocpole(vysledek.sirka, vysledek.vyska);

  // nacitam pole:

  for(i=0; i<vysledek.vyska; i++){
    mf_najdi(soubor, '"');
    for(j=0; j<vysledek.sirka; j++){
      znak = mf_znak(soubor);
      for(pom=prvbarva; pom; pom=pom->next)
	if(znak && znak == pom->znak){
	  vysledek.pole[j + i*vysledek.sirka] = pom;
	  break;
	}
      if(pom == NULL){
	warning ("Neznama barva: %d='%c'", znak, znak);
	mf_error(soubor);
      }
    }
    mf_musibyt(soubor, "\"");
  }
  mf_zavri(soubor);
  free(fullname);

  return vysledek;
}

kq_submap *nejvyssipix (int x, int y)
{
  kq_submap *pom, *max;
  kq_color *c;
  float z;

  max = NULL;
  for(pom = prvsubmap; pom; pom = pom->next){
    c = smpixel(pom, x,y);
    if(c && (max == NULL || c->z+pom->z > z)){
      z = c->z+pom->z;
      max = pom;
    }
  }

  return max;
}

static void pouzijsubmap (kq_submap *sm)
{
  int x, y, sx, sy;

  if(!mapa_z) return;

  if(sm->x < 0){
    x = 0;
    sx = -sm->x;
  }
  else{
    x = sm->x;
    sx = 0;
  }
  for(; sx < sm->map.sirka && x < mapa.sirka; x++, sx++){
    if(sm->y < 0){
      y = 0;
      sy = -sm->y*sm->map.sirka;
    }
    else{
      y = sm->y*mapa.sirka;
      sy = 0;
    }
    for(; sy < sm->map.vyska*sm->map.sirka && y < mapa.vyska*mapa.sirka;
	y += mapa.sirka, sy += sm->map.sirka){
      if(sm->z + sm->map.pole[sx + sy]->z >= mapa_z[x + y]){
	mapa_z[x + y] = sm->z + sm->map.pole[sx + sy]->z;
	mapa.pole[x + y] = sm->map.pole[sx + sy];
      }
    }
  }
}

static void zrussubmap (kq_submap *sm)
{
  int x, y, sx, sy;
  kq_submap *pom;

  if(!mapa_z) return;

  if(sm->x < 0){
    x = 0;
    sx = -sm->x;
  }
  else{
    x = sm->x;
    sx = 0;
  }
  for(; sx < sm->map.sirka && x < mapa.sirka; x++, sx++){
    if(sm->y < 0){
      y = 0;
      sy = -sm->y*sm->map.sirka;
    }
    else{
      y = sm->y*mapa.sirka;
      sy = 0;
    }
    for(; sy < sm->map.vyska*sm->map.sirka && y < mapa.vyska*mapa.sirka;
	y += mapa.sirka, sy += sm->map.sirka){
      if(mapa_z[x + y] == sm->map.pole[sx + sy]->z + sm->z &&
	 mapa.pole[x + y] == sm->map.pole[sx + sy]){
	pom = nejvyssipix(x, y/mapa.sirka);
	mapa.pole[x + y] = smpixel(pom, x, y/mapa.sirka);
	mapa_z[x + y] = mapa.pole[x + y]->z + pom->z;
      }
    }
  }
}

static void zaradsubmap (kq_submap *sm, kq_submap **prvni)
{
  while(*prvni && (*prvni)->z > sm->z) prvni = &(*prvni)->next;

  sm->next = *prvni;
  *prvni = sm;
}

static void vyradsubmap (kq_submap *sm, kq_submap **prvni)
{
  while(*prvni && *prvni != sm) prvni = &(*prvni)->next;
  if(!*prvni){
    warning("submapa nenalezena");
    return;
  }

  *prvni = sm->next;
}

void zmenbgmap (char *filename)
{
  if(bgmap){
    free(bgmap->map.pole);
    free(bgmap->nazev);
  }

  if(!filename){
    if(bgmap){
      vyradsubmap(bgmap, &prvsubmap);
      free(bgmap);
      bgmap = NULL;
    }
    return;
  }

  if(!bgmap){
    bgmap = (kq_submap *)malloc(sizeof(kq_submap));
    bgmap->x = 0;
    bgmap->y = 0;
    bgmap->z = 0;
    bgmap->hide = 0;
    bgmap->lokalni = 0;
    zaradsubmap(bgmap, &prvsubmap);
  }

  bgmap->map = nactimapu(filename);
  bgmap->nazev = strdup(filename);
  if(autoreloadmap) reloadmap();
}

kq_submap *vytvorsubmap (int x, int y, float z, char *filename,
		       char lokalni, char hide)
{
  kq_submap *vysledek;

  vysledek = (kq_submap *)malloc(sizeof(kq_submap));;
  vysledek->x = x;
  vysledek->y = y;
  vysledek->z = z;
  vysledek->hide = hide;
  vysledek->lokalni = lokalni;
  vysledek->map = nactimapu(filename);
  vysledek->nazev = strdup(filename);

  if(hide) zaradsubmap(vysledek, &prvhidsm);
  else{
    zaradsubmap(vysledek, &prvsubmap);
    if(autoreloadmap) pouzijsubmap(vysledek);
  }

  return vysledek;
}

void reloadmap ()
{
  int i;
  kq_submap *pom;

  if(!bgmap){
    if(mapa_z){
      free(mapa_z);
      free(mapa.pole);
    }
    mapa.pole = NULL;
    mapa_z = NULL;
    return;
  }

  if(!mapa_z ||
     mapa.sirka != bgmap->map.sirka ||
     mapa.vyska != bgmap->map.vyska){

    if(mapa_z){
      free(mapa_z);
      free(mapa.pole);
    }

    mapa.sirka = bgmap->map.sirka;
    mapa.vyska = bgmap->map.vyska;
    mapa.pole = allocpole(mapa.sirka, mapa.vyska);
    mapa_z = (float *)malloc(sizeof(float *)*mapa.sirka*mapa.vyska);
  }

  for(i=0; i<mapa.sirka*mapa.vyska; i++){
    mapa.pole[i] = bgmap->map.pole[i];
    mapa_z[i] = bgmap->map.pole[i]->z;
  }

  for(pom = prvsubmap; pom; pom = pom->next)
    if(pom != bgmap) pouzijsubmap(pom);
}

kq_color *mappixel (kq_map map, int x, int y)
{
  if(!map.pole) return NULL;
  if(x < 0 || y < 0 || x >= map.sirka || y >= map.vyska) return NULL;
  return map.pole[x+y*map.sirka];
}

kq_color *smpixel (kq_submap *sm, int x, int y)
{
  return mappixel(sm->map, x-sm->x, y-sm->y);
}

kq_submap *najdibarvu (kq_color *col, int x, int y)
{
  kq_submap *pom;

  for(pom = prvsubmap; pom; pom = pom->next)
    if(col == smpixel(pom, x,y)) return pom;

  return NULL;
}

void hidesubmap (kq_submap *sm)
{
  if(sm == bgmap){
    warning("hidesubmap(bgmap)");
    return;
  }

  if(sm->hide) return;

  vyradsubmap(sm, &prvsubmap);
  if(autoreloadmap) zrussubmap(sm);

  sm->hide = 1;

  sm->next = prvhidsm;
  prvhidsm = sm;
}

void unhidesubmap (kq_submap *sm)
{
  if(sm == bgmap){
    warning("unhidesubmap(bgmap)");
    return;
  }

  if(!sm->hide) return;

  vyradsubmap(sm, &prvhidsm);
  
  sm->hide = 0;

  zaradsubmap(sm, &prvsubmap);
  if(autoreloadmap) pouzijsubmap(sm);
}

static void freesm (kq_submap *sm)
{
  free(sm->map.pole);
  free(sm->nazev);
  free(sm);
}

void znicsubmap (kq_submap *sm)
{
  if(sm == bgmap){
    warning("znicsubmap(bgmap)");
    return;
  }

  if(sm->hide) vyradsubmap(sm, &prvhidsm);
  else{
    vyradsubmap(sm, &prvsubmap);
    if(autoreloadmap) zrussubmap(sm);
  }

  freesm(sm);
}

void deletelsubmaps ()
{
  kq_submap **pom, *pomo, *pomoc;
  char trebarel;

  if(bgmap -> lokalni){
    warning("deletelsubmaps: lokalni bgmap");
    bgmap->lokalni = 0;
  }

  for(pom = &prvhidsm;; pom = &pomo->next){
    for(pomo = *pom; pomo && pomo->lokalni > 0; pomo = pomoc){
      pomoc = pomo->next;
      freesm(pomo);
    }
    *pom = pomo;
    if(!pomo) break;
  }

  trebarel = 0;
  for(pom = &prvsubmap;; pom = &pomo->next){
    for(pomo = *pom; pomo && pomo->lokalni; pomo = pomoc){
      pomoc = pomo->next;

      trebarel = 1;

      if(pomo->lokalni > 0) freesm(pomo);
      else{
	pomo->next = prvhidsm;
	prvhidsm = pomo;
      }
    }
    *pom = pomo;
    if(!pomo) break;
  }

  if(autoreloadmap && trebarel) reloadmap();
}

void deleteAsubmaps ()
{
  kq_submap *pom;

  while(prvsubmap){
    pom = prvsubmap->next;
    freesm(prvsubmap);
    prvsubmap = pom;
  }
  while(prvhidsm){
    pom = prvhidsm->next;
    freesm(prvhidsm);
    prvhidsm = pom;
  }

  bgmap = NULL;
  if(autoreloadmap){
    free(mapa_z);
    free(mapa.pole);
    mapa_z = NULL;
    mapa.pole = NULL;
  }
}
