#include<stdio.h>
#include<stdlib.h>
#include "types.h"
#include "pin.h"
#include "warning.h"
#include "timer.h"

/*
  Timery jsou ulozeny v seznamu od toho, co se ma provest nejdrive (prvtimer)
  po ten, co se ma provest nejpozdeji. V promenne zbyva struktury timer je
  rozdil casu, za ktery se ma provest dany timer a predchozi timer. V kazdem
  framu tedy staci odecist jednicku v promenne zbyva prvniho timeru. Funkce
  zaradtimer() zaradi timer do seznamu timeru podle casu, za ktery se ma
  provest. Krome casu pri zarazovani zohlednuje i prioritu.
*/

//: aktutimer

kq_timer *aktutimer = NULL;

//: seznamy

kq_timer *prvtimer = NULL;
kq_timer *pausedtimers = NULL;

//: zaradvyrad

static void zaradtimer (kq_timer *tim, int cas)
/*********************************************/
{
  kq_timer **pom;

  if(tim->pause){
    tim->zbyva = cas;
    tim->next = pausedtimers;
    pausedtimers = tim;
    return;
  }

  for(pom = &prvtimer;; pom = &(*pom)->next){ // vyhledam vhodne misto
    if(!*pom) break;
    cas -= (*pom)->zbyva;  // cas se nakonec ulozi do zbyva
    if(cas < 0) break;
    if(cas == 0 && tim->priorita >= (*pom)->priorita) break;
  }

  if(*pom){
    cas += (*pom)->zbyva; // cas jsem prestrelil
    (*pom)->zbyva -= cas; // upravim zbyva v nasledujicim timeru
  }

  tim->zbyva = cas;
  tim->next = *pom;
  *pom = tim;
}

static void vyradtimer (kq_timer *tim)
/************************************/
{
  kq_timer **pom;

  if(tim->pause) pom = &pausedtimers;
  else pom = &prvtimer;

  for(; *pom != tim; pom = &(*pom)->next)
    if(!*pom){
      warning("vyradtimer(): timer nenalezen");
      return;
    }

  *pom = tim->next;
  if(tim->next && !tim->pause) tim->next->zbyva += tim->zbyva;
}

//: vytvortimer

kq_timer *vytvortimer (void (*funkce)(), kq_pin parametr, int cas,
		       float priorita, int opakovani, char lokalni, char pause)
/************************************************************************/
{
  kq_timer *tim;
  
  tim = (kq_timer *)malloc(sizeof(kq_timer));
  tim->funkce = funkce;
  tim->parametr = parametr;

  tim->cas = cas;
  tim->priorita = priorita;
  tim->opakovani = opakovani;
  tim->lokalni = lokalni;
  tim->pause = pause;

  zaradtimer(tim, tim->cas);

  return tim;
}

//: zjistitimer

int zjistitimer (kq_timer *tim)
/*****************************/
{
  int cas;
  kq_timer *pom;

  if(tim->pause || tim == aktutimer) return tim->zbyva;

  cas=0;

  for(pom = prvtimer; tim != pom; pom = pom->next){
    if(!pom){
      warning("zjistitimer: timer nenalezen\n");
      return -1;
    }
    cas += pom->zbyva;
  }
  return cas + tim->zbyva;
}

//: zmentimer

void zmentimer (kq_timer *tim, int cas, float priorita)
/*****************************************************/
{
  tim->priorita = priorita;
  if(tim->pause || tim == aktutimer)
    tim->zbyva = cas;
  else{
    vyradtimer(tim);
    zaradtimer(tim, cas);
  }
}

//: pause

void pausetimer (kq_timer *tim)
/********************************/
{
  kq_timer **pom;

  if(tim->pause) return;
  tim->pause = 1;
  if(tim == aktutimer) return;

  if(tim->next) tim->next->zbyva += tim->zbyva;

  for(pom = &prvtimer; *pom != tim; pom = &(*pom)->next){
    if(!*pom){
      warning("pausetimer(): timer nenalezen\n");
      return;
    }
    tim->zbyva += (*pom)->zbyva;
  }

  *pom = tim->next;
  tim->next = pausedtimers;
  pausedtimers = tim;
}

void unpausetimer(kq_timer *tim)
/********************************/
{
  if(!tim->pause) return;
  if(tim == aktutimer){
    tim->pause = 0;
    return;
  }
  vyradtimer(tim);
  tim->pause = 0;
  zaradtimer(tim, tim->zbyva);
}

//: znictimer

void znictimer(kq_timer *tim)
/***************************/
{
  if(tim == aktutimer) aktutimer = NULL;
  else vyradtimer(tim);
  free(tim);
}

//: deleteltimers

void deleteltimers()
/******************/
{
  kq_timer **pom, *pomo, *pomoc;
  int i;

  // Prochazim zapausovane timery:

  for(pom = &pausedtimers;; pom = &pomo->next){
    for(pomo = *pom; pomo && pomo->lokalni > 0; pomo = pomoc){
      pomoc = pomo->next;
      free(pomo);
    }
    *pom = pomo;
    if(!pomo) break;
  }
 
  // Prochazim spustene timery.

  i=0;
  for(pom = &prvtimer; (pomo = *pom); pom = &pomo->next){
    i += pomo->zbyva;
    for(; pomo && pomo->lokalni; pomo = pomoc){
      pomoc = pomo->next;
      if(pomoc){
	i += pomoc->zbyva;
	pomoc->zbyva += pomo->zbyva;
	pomo->zbyva += i - pomoc->zbyva;
      }
      else pomo->zbyva=i;

      if(pomo->lokalni > 0) free(pomo);
      else{
	pomo->pause = 1;
	pomo->next = pausedtimers;
	pausedtimers = pomo;
      }
    }
    *pom = pomo;
    if(!pomo) break;
  }

  // Zkontroluji aktualni timer:

  if(aktutimer && aktutimer->lokalni){
    if(aktutimer->lokalni < 0) aktutimer->pause = 1;
    else{
      free(aktutimer);
      aktutimer = NULL;
    }
  }
}

//: deleteAtimers

void deleteAtimers()
/******************/
{
  kq_timer *pom;

  while(prvtimer){
    pom = prvtimer->next;
    free(prvtimer);
    prvtimer = pom;
  }
  while(pausedtimers){
    pom = pausedtimers->next;
    free(pausedtimers);
    pausedtimers = pom;
  }
  if(aktutimer){
    free(aktutimer);
    aktutimer = NULL;
  }
}

//: dotimers

void dotimers()
/*************/
{
  while(prvtimer && !prvtimer->zbyva){
    // Vyradim timer ze seznamu:

    aktutimer = prvtimer;
    prvtimer = prvtimer->next;

    aktutimer->zbyva = aktutimer->cas;
    aktutimer->next = NULL;
    if(aktutimer->opakovani >= 0) aktutimer->opakovani--;

    // Spustim ho:

    spustsparem(aktutimer->funkce, &aktutimer->parametr);

    // Vratim ho zpatky nebo smazu:

    if(aktutimer){ // timer se mohl sam smazat
      if(!aktutimer->opakovani)
	free(aktutimer); // Timer uz se provedl, kolikrat mel.
      else
	zaradtimer(aktutimer, aktutimer->zbyva);
    }
  }
  aktutimer = NULL;
}

//: ladici_tisky

void vypistimer(kq_timer *tim)
{
  if(tim == NULL){
    printf("NULL\n");
    return;
  }
  printf("- %d -> %d x ", tim->zbyva, tim->opakovani);
  if(tim->lokalni > 0) printf("local ");
  if(tim->lokalni == 0) printf("global ");
  printf("timer: cas=%d, prior.=%f, funkce(", tim->cas, tim->priorita);
  if(tim->parametr.typ == PININT) printf("%d", tim->parametr.u.i);
  if(tim->parametr.typ == PINPOINTER) printf("pointer");
  printf(")");
  if(tim->pause) printf("\tPAUSED");
  printf("\n");
}

void vypistimery()
{
  int i, j;
  kq_timer *pom;

  printf("aktutimer:\n  ");
  vypistimer(aktutimer);
  printf("pausedtimers:\n");
  for(pom = pausedtimers; pom; pom = pom->next){
    printf("  ");
    vypistimer(pom);
  }
  printf("timers:\n");
  i=1;
  for(pom = prvtimer; pom; pom = pom->next){
    for(j=0; j<i; j++) printf("  ");
    vypistimer(pom);
    i++;
  }
  printf("\n");
}
